/*
 * 入力が途心10マーガレットのミニマックス解であるかどうかを判定する
 *
 * 入力には、以下の二つの仮定を置く
 *
 *  要求1:  HP10以下でダメージソース(鏡を含む)を持たない剣師は全て勝率0.5未満
 *  要求2:  5/0/0/1/勝3ABC/の形の剣師は全て勝率0.5未満(A、B、Cはマーガレットの任意の剣)
 *
 * 要求2はcheck_5001関数で検査を行う
 * 要求1についてはこのコードでは検査をしないので、利用者が検証すること
 */

/*
 * 探索の省略について
 *
 * 仮定1: M≧Nであるとき、剣師xの回復剣Mを勝利剣Nに置換した剣師はxの上位互換
 * 仮定2: 身体能力5/0/0/1/の剣師は、5/0/0/1/勝3ABC/という形の剣師いずれかの下位互換
 *
 * これらは共に証明の必要な命題だが、たぶん正しいだろう
 *
 * まず、要求2と仮定2から、身体能力5/0/0/1/の剣師は入力環境下で勝率0.5未満のため、探索を省略する
 *
 * 探索の際、LIMIT_RECOVERYを超える遅延数の回復剣("遅い"回復剣)はひとまず考慮せず、代わりに遅延起動(LIMIT_RECOVERY+1)の勝利剣を探索範囲に含める
 * 勝利剣を持つ剣師の勝率が0.5未満であれば、仮定1から、その剣師の勝利剣を遅い回復剣に置換してできる剣師の勝率も0.5未満であるので、探索を省略する
 * 勝利剣を持つ剣師の勝率が0.5以上だった場合、このショートカットは使えないので、それぞれの勝利剣を遅い回復剣に置換してできる剣師をすべて探索する
 *
 * 要求1から、HP10以下でダメージソースを持たない剣師については探索を省略する
 */

/*
 * 上のものに加えて、search.cが以下の暗黙の仮定を置いている
 *
 * 仮定3: あいざわである剣師は、それよりボーナスが少なくあいざわでない剣師の一つ以上と同値
 * 仮定4: 回M回Nの並びを持つ剣師は、回N回Mと並べ換えても同値
 * 仮定5: 0-1ターン目に利用可能なリミテッド剣と、それより遅延の大きい剣が隣接していたら、それらを並べ換えても同値
 */

# define LIMIT_RECOVERY 6 /* 通常の探索をする回復剣の遅延上限 */
# define SEARCH_HIGH_RECOVERY_OWNER /* 遅延がLIMIT_RECOVERYを超える回復剣を持つ剣師を探索する */
# define SEARCH_HIGH_RECOVERY_NONOWNER /* 遅延がLIMIT_RECOVERYを超える回復剣を持たない剣師を探索する */

# define THE_VICTORY S_VICTORY(LIMIT_RECOVERY+1)

# include <stdio.h>
# include <stdlib.h>
# include <time.h>
# include <gmp.h>
# include "search.h"

# include "parallelizer.h"

static int max(int x, int y) { return x < y ? y : x; }
static int min(int x, int y) { return x < y ? x : y; }

void print_kenshi_with(const char *prefix, const Kenshi *k)
{
  char buf[1000];
  ken_to_theki(k, buf);
  printf("%s%s\n", prefix, buf);
}

void error_exit(const char *msg)
{
  fprintf(stderr, "error: %s\n", msg);
  exit(1);
}

typedef struct
{
  Kenshi *ks;
  mpz_t *ws;
  int num_group;
  int *group_indices;
  mpz_t total;
} env;

void read_env(env *out)
{
  int i, k;
  Kenshi *ks;
  mpz_t *ws;
  int num_kenshi;
  int *group_indices;
  if(scanf("%d", &num_kenshi) != 1)
    error_exit("failed to read env size");
  ks = (Kenshi *)malloc(sizeof(Kenshi) * num_kenshi);
  ws = (mpz_t *)malloc(sizeof(mpz_t) * num_kenshi);
  group_indices = (int *)malloc(sizeof(int) * num_kenshi + 1);
  mpz_init(out->total);
  i = k = 0;
  while(i < num_kenshi)
  {
    char *p, *end;
    char buf[1001];
    if(!mpz_inp_str(ws[k], stdin, 10))
      error_exit("weight");
    mpz_add(out->total, out->total, ws[k]);
    group_indices[k] = i;
    if(!fgets(buf, 1001, stdin))
      error_exit("spec");
    for(p = buf; *p; )
    {
      for(end = p; *end != '\n' && *end != ','; end++) ;
      *end = '\0';
      ken_read_theki(&ks[i++], p);
      p = end + 1;
    }
    k++;
  }
  group_indices[k] = i;
  out->ks = ks;
  out->ws = ws;
  out->num_group = k;
  out->group_indices = group_indices;
  printf("L total = ");
  mpz_out_str(stdout, 10, out->total);
  printf("\n");
}

void destroy_env(env *e)
{
  int i;
  for(i = 0; i < e->num_group; i++)
    mpz_clear(e->ws[i]);
  free(e->ws);
  free(e->ks);
  free(e->group_indices);
  mpz_clear(e->total);
}

void show_env(const env *e)
{
  int i, j;

  for(i = 0; i < e->num_group; i++)
  {
    printf("L group %d:\n", i);
    for(j = e->group_indices[i]; j < e->group_indices[i+1]; j++)
      print_kenshi_with("L   ", &e->ks[j]);
  }
}

/* eの第iグループの剣師をxと戦わせ、xにとって最善の結果を返す */
int match_against_group(const Kenshi *x, const env *e, int i)
{
  int k;
  int drawn = 0;
  for(k = e->group_indices[i]; k < e->group_indices[i+1]; k++)
    switch(bd_go_by_cpy(x, &e->ks[k]))
    {
    case C1_WIN:
      return C1_WIN;
    case DRAW:
      drawn = 1;
      break;
    default:
      break;
    }
  return drawn ? DRAW : C2_WIN;
}

/* 環境eにおけるxの勝率を0.5と比較する。勝率が0.5を超えるなら1を、下回るなら-1を、ちょうど0.5なら0を返す */
int test_kenshi(const Kenshi *x, const env *e)
{
  int i;
  mpz_t current;
  int result;
  mpz_init_set(current, e->total);

  for(i = 0; i < e->num_group; i++)
  {
    switch(match_against_group(x, e, i))
    {
    case C2_WIN:
    {
      mpz_t temp;
      mpz_init(temp);
      mpz_mul_2exp(temp, e->ws[i], 1);
      mpz_sub(current, current, temp);
      mpz_clear(temp);
      break;
    }
    case DRAW:
      mpz_sub(current, current, e->ws[i]);
      break;
    default:
      break;
    }
    if(mpz_sgn(current) < 0)
      break;
  }
  result = mpz_sgn(current);
  mpz_clear(current);
  return result;
}

typedef struct
{
  const env *e;
  time_t last_showed;
  int counter;
  int expand_counter;
  int id;
} context;

void init_context(context *ctx, const env *e, int id)
{
  ctx->e = e;
  ctx->last_showed = time(NULL);
  ctx->counter = 0;
  ctx->expand_counter = 0;
  ctx->id = id;
}

void tick(context *ctx, const Kenshi *x)
{
  time_t now = time(NULL);
  if(now != ctx->last_showed)
  {
    char buf[100];
    sprintf(buf, "L[%d] ", ctx->id);
    print_kenshi_with(buf, x);
    ctx->last_showed = now;
  }
  ctx->counter++;
}

void test_and_print(const Kenshi *x, context *ctx)
{
  tick(ctx, x);
  switch(test_kenshi(x, ctx->e))
  {
  case 1:
    print_kenshi_with("+++ ", x);
    break;
  case 0:
    print_kenshi_with("= ", x);
    break;
  default:
    break;
  }
}

void expand(const Kenshi *orig, context *ctx, int *swords, const int *victories, int pos)
{
  if(victories[pos] == -1)
  {
    Kenshi x;
    ken_init(&x, orig->hp, orig->attack, orig->defense, orig->speed, swords);
    test_and_print(&x, ctx);
  }
  else
  {
    int i;
    for(i = LIMIT_RECOVERY + 1; i <= 20; i++)
    {
      swords[victories[pos]] = S_RECOVERY(i);
      expand(orig, ctx, swords, victories, pos + 1);
    }
  }
}

/* 勝利剣を持つ剣師を入力とし、その勝利剣を遅い回復剣に置換した剣師で、
 * eにおいて勝率が0.5以上であるものを全て表示する
 */
void test_and_expand(const Kenshi *x, context *ctx)
{
  int i;
  int victories[10];
  int n_victory;
  int swords[SWORD_SIZE];

  tick(ctx, x);

  if(test_kenshi(x, ctx->e) == -1) return; /* 勝利剣を持っても勝率0.5に満たないなら、回復剣に持ち替えても満たない */

  ctx->expand_counter++;

  n_victory = 0;
  for(i = 0; x->swords[i] != -1; i++)
  {
    swords[i] = x->swords[i];
    if(swords[i] == THE_VICTORY)
      victories[n_victory++] = i;
  }
  swords[i] = -1;
  victories[n_victory] = -1;

  expand(x, ctx, swords, victories, 0);
}

void enum_all(int spd, int has_hp, const int *swords, int n_sword, int bonus, void *param)
{
  int i;
  int attack;
  int sword_count[SWORD_VARIATION] = {0};
  int has_attack_independent, has_attack_dependent;
  for(i = 0; i < n_sword; i++)
    sword_count[swords[i]]++;
# ifndef SEARCH_HIGH_RECOVERY_NONOWNER
  if(!sword_count[THE_VICTORY])
    return;
# endif
  has_attack_independent = sword_count[S_MAGICAL] || sword_count[S_DEATH];
  has_attack_dependent = sword_count[S_SLASH] || sword_count[S_RAPID] || sword_count[S_HEAVY] ||
    sword_count[S_SHOCK] || sword_count[S_POISON] | sword_count[S_ABSOLUTE];
  for(attack = 0; attack <= bonus; attack++)
  {
    int hp_max = bonus-attack, hp_min = 1;
    int hp;
    if(attack > 0 && !has_attack_dependent) continue;
    if(!has_hp)
      hp_max = hp_min = 0;
    for(hp = hp_min; hp <= hp_max; hp++)
    {
      Kenshi x;
      int defense = bonus-attack-hp;

      if(hp <= 2 && !has_attack_independent && (!has_attack_dependent || !attack && !sword_count[S_HEAT]) && !sword_count[S_MIRROR])
        continue; /* HP10以下でダメージソースがない */
      ken_init(&x, 5*hp, attack, defense, spd, swords);
      if(sword_count[THE_VICTORY])
        test_and_expand(&x, (context *)param);
      else
        test_and_print(&x, (context *)param);
      
      /*
      Kenshi x;
      int defense = bonus-attack-hp;
      ken_init(&x, 5*hp, attack, defense, spd, swords);
      print_kenshi(&x);
      */
    }
  }
}

typedef struct
{
  int spd;
  int has_hp;
  int n_sword;
  int bonus;
  const int *sword_env;
  context ctx;
} task;

void perform_task(void *param)
{
  task *t = (task *)param;
  each_sword_array(t->spd, t->has_hp, t->n_sword, t->bonus, t->sword_env, enum_all, &t->ctx);
  free(t);
}

void search(const int *swords, const env *set)
{
  int cid = 0;
  parallelizer *pz = new_pz();
  int bonus;
  for(bonus = 10; bonus >= 0; bonus--)
  {
    int has_hp;
    printf("L bonus: %d\n", bonus);

    for(has_hp = 0; has_hp <= 1; has_hp++)
    {
      int spd;
      for(spd = bonus-has_hp; spd >= 0; spd--)
      {
        task *t = (task *)malloc(sizeof(task));
        t->spd = spd;
        t->has_hp = has_hp;
        t->n_sword = bonus-spd-has_hp-(has_hp&&spd==1?1:0);
        t->bonus = bonus-spd;
        t->sword_env = swords;
        init_context(&t->ctx, set, cid++);

        do_pz(pz, perform_task, t);
      }
    }
  }
  wait_and_destroy_pz(pz);
}

void check_main(const env *e)
{
  int swords[100];

# ifdef SEARCH_HIGH_RECOVERY_OWNER
  swords[0] = S_VICTORY(LIMIT_RECOVERY+1);
  set_swords_within(swords+1, LIMIT_RECOVERY);
# else
  set_swords_within(swords, LIMIT_RECOVERY);
# endif

  search(swords, e);
}

/* 全ての5/0/0/1/勝3ABC/の形の剣師がeにおいて勝率0.5未満であることを確かめる */
void check_5001(const env *e)
{
  int i, j, k;
  for(i = 0; i <= S_RECOVERY(21); i++)
    for(j = 0; j <= S_RECOVERY(21); j++)
      for(k = 0; k <= S_RECOVERY(21); k++)
      {
        int swords[5];
        Kenshi x;
        swords[0] = S_VICTORY(3);
        swords[1] = i;
        swords[2] = j;
        swords[3] = k;
        swords[4] = -1;
        ken_init(&x, 5, 0, 0, 1, swords);
        if(test_kenshi(&x, e) >= 0)
        {
          print_kenshi_with("5001 check failed at: ", &x);
          exit(1);
        }
      }
}

int main(void)
{
  env e;

  setvbuf(stdout, NULL, _IOLBF, 0);

  read_env(&e);
  show_env(&e);

  check_5001(&e);
  check_main(&e);

  destroy_env(&e);

  return 0;
}
